import numpy as np
from scipy.stats import norm
from .boundsWMWStatistic import bounds_wmw_statistic
from .exactDistributionWMWStatistic import dis_WMW_exact


def bounds_p_value_no_ties(X, Y, alternative, exact=None, correct=False):
    # Sample size
    n = len(X)
    m = len(Y)

    # Compute bounds of WMWS statistic using bounds_wmw_statistic function
    Bounds = bounds_wmw_statistic(X, Y, ties=False, lower_boundary=-np.inf, upper_boundary=np.inf)
    lowerBoundWMWStatistic = Bounds[0]
    upperBoundWMWStatistic = Bounds[1]

    # Decide if exact 
    if exact is None:
        exact = (n < 50) and (m < 50)

    # Compute p_1, p_2 -- the p-values corresponding to the minimum and maximum WMW test statistics, respectively
    if exact:
        # compute p_1
        if alternative == 'two.sided':
            if lowerBoundWMWStatistic > (n * m / 2):
                p = dis_WMW_exact(lowerBoundWMWStatistic-1, n, m, lower_tail=False)
            else:
                p = dis_WMW_exact(lowerBoundWMWStatistic, n, m)
            p_1 = min(2 * p, 1)
        elif alternative == 'greater':
            p_1 = dis_WMW_exact(lowerBoundWMWStatistic-1, n, m, lower_tail=False)
        elif alternative == 'less':
            p_1 = dis_WMW_exact(lowerBoundWMWStatistic, n, m)
        
        # compute p_2
        if alternative == 'two.sided':
            if lowerBoundWMWStatistic > (n * m / 2):
                p = dis_WMW_exact(upperBoundWMWStatistic-1, n, m, lower_tail=False)
            else:
                p = dis_WMW_exact(upperBoundWMWStatistic, n, m)
            p_2 = min(2 * p, 1)
        elif alternative == 'greater':
            p_2 = dis_WMW_exact(upperBoundWMWStatistic-1, n, m, lower_tail=False)
        elif alternative == 'less':
            p_2 = dis_WMW_exact(upperBoundWMWStatistic, n, m)
            
    else:
        # if not exact, using normal approximation
        correct_lower = 0
        correct_upper = 0

        if correct:
            # using continuity correction
            if alternative == "two.sided":
                correct_lower = np.sign(lowerBoundWMWStatistic - n * m / 2) * 0.5
                correct_upper = np.sign(upperBoundWMWStatistic - n * m / 2) * 0.5
            elif alternative == "greater":
                correct_lower = 0.5
                correct_upper = 0.5
            elif alternative == "less":
                correct_lower = -0.5
                correct_upper = -0.5

        mu = n * m / 2
        sigma = np.sqrt(n * m * (n + m + 1) / 12)

        lowerBoundZ = (lowerBoundWMWStatistic - correct_lower - mu) / sigma
        upperBoundZ = (upperBoundWMWStatistic - correct_upper - mu) / sigma

        if alternative == 'two.sided':
            if lowerBoundZ < 0:
                p = norm.cdf(lowerBoundZ)
            else:
                p = 1 - norm.cdf(lowerBoundZ)
            p_1 = 2 * p
        elif alternative == 'greater':
            p_1 = norm.sf(lowerBoundZ)
        elif alternative == 'less':
            p_1 = norm.cdf(lowerBoundZ)

        if alternative == 'two.sided':
            if upperBoundZ < 0:
                p = norm.cdf(upperBoundZ)
            else:
                p = 1 - norm.cdf(upperBoundZ)
            p_2 = 2 * p
        elif alternative == 'greater':
            p_2 = norm.sf(upperBoundZ)
        elif alternative == 'less':
            p_2 = norm.cdf(upperBoundZ)

    # Compute bounds of p-values
    lowerBoundPValue = min(p_1, p_2)

    if alternative == 'two.sided':
        if ((lowerBoundWMWStatistic - n * m / 2) * (upperBoundWMWStatistic - n * m / 2)) > 0:
            upperBoundPValue = max(p_1, p_2)
        else:
            upperBoundPValue = 1
    elif alternative == 'greater':
        upperBoundPValue = p_1
    elif alternative == 'less':
        upperBoundPValue = p_2

    return np.array([lowerBoundWMWStatistic, upperBoundWMWStatistic, lowerBoundPValue, upperBoundPValue, exact])

# Example usage:
# X = np.array([1, 2, np.nan, 4, 5])
# Y = np.array([3, np.nan, 6, 7])
# alternative = 'two.sided'
# exact = None
# correct = False
# result = bounds_p_value_no_ties(X, Y, alternative, exact, correct)
# print("Lower bound of WMW Statistic:", result[0])
# print("Upper bound of WMW Statistic:", result[1])
# print("Lower bound of p-value:", result[2])
# print("Upper bound of p-value:", result[3])
# print("Exact:", result[4])



import numpy as np
from scipy.stats import rankdata

def bounds_sum_rank(X, Y, ties, lower_boundary, upper_boundary):
    
    # Sample size
    n = len(X)
    m = len(Y)

    # Observed samples in X and Y, respectively
    X_prime = X[~np.isnan(X)]
    Y_prime = Y[~np.isnan(Y)]

    # All observed samples
    Z_prime = np.concatenate((X_prime, Y_prime))

    # Observed sample size
    n_prime = len(X_prime)
    m_prime = len(Y_prime)

    # Sum of ranks of X_prime in Z_prime
    rankSumX_prime = np.sum(rankdata(Z_prime, method = 'average')[:n_prime])

    # Bounds of sum of ranks of X in Z without ties
    lowerBoundSumRank = (rankSumX_prime + 
                         (n - n_prime) * (n + n_prime + 1) / 2)
    
    upperBoundSumRank = (rankSumX_prime +
                         (n * (n + 2 * m + 1) - n_prime * 
                          (n_prime + 2 * m_prime + 1)) / 2)

    # Bounds of sum of ranks of X in Z with ties
    if ties:
        # If ties are allowed, tighter bounds may exist
        min_observed = np.min(Z_prime)
        max_observed = np.max(Z_prime)

        if lower_boundary <= min_observed:
            a = lower_boundary
        else:
            print('lower_boundary must be smaller or equal than the minimum of all observed data. lower_boundary is set to -Inf')
            a = -np.inf

        if upper_boundary >= max_observed:
            b = upper_boundary
        else:
            print('upper_boundary must be larger or equal than the maximum of all observed data. upper_boundary is set to Inf')
            b = np.inf

        lowerBoundSumRank += (np.sum(np.array(Y_prime) == a) * (n - n_prime) \
                              + np.sum(np.array(X_prime) == b) \
                              * (m - m_prime)) / 2

        upperBoundSumRank -= (np.sum(np.array(X_prime) == a) * (m - m_prime) \
                              + np.sum(np.array(Y_prime) == b) \
                              * (n - n_prime)) / 2

    return np.array([lowerBoundSumRank, upperBoundSumRank])

# Example usage:
# X = np.array([1, 2, np.nan, 4, 5])
# Y = np.array([3, np.nan, 6, 7])
# ties = True
# lower_boundary = -np.inf
# upper_boundary = np.inf

# result = bounds_sum_rank(X, Y, ties, lower_boundary, upper_boundary)
# print("Lower bound:", result[0])
# print("Upper bound:", result[1])





import numpy as np
from .boundsSumRank import bounds_sum_rank


def bounds_wmw_statistic(X, Y, ties, lower_boundary, upper_boundary):
    # Sample size
    n = len(X)

    # Compute bounds of sum of ranks using bounds_sum_rank function
    Bounds = bounds_sum_rank(X, Y, ties, lower_boundary, upper_boundary)
    lowerBoundSumRank = Bounds[0]
    upperBoundSumRank = Bounds[1]

    # Compute bounds of WMW test statistic
    lowerBoundWMWStatistic = lowerBoundSumRank - n * (n + 1) / 2
    upperBoundWMWStatistic = upperBoundSumRank - n * (n + 1) / 2

    return np.array([lowerBoundWMWStatistic, upperBoundWMWStatistic])



# Compute the exact distribution of WMW test statistic

import numpy as np
import math

def num_seq(U, n, m):
    '''
    Given the sample sizes n, m of x and y, return the number of sequences such that the values of x precedes the values of y U times. For more details, see the reference provided.
    
    Parameters
    ----------
    U : int
        Value for which the number is computed.
    n : int
        Sample size of x.
    m : int
        Sample size of y.
        
    Returns
    -------
    int
        The number of sequences such that the values from x precedes the values from y U times
    
    Reference
    ----------
    Mann HB, Whitney DR. On a test of whether one of two random variables is stochastically larger than the other. The annals of mathematical statistics. 1947 Mar 1:50-60.

    '''

    if U < 0:
        return 0

    if (n == 0 or m == 0) and U != 0:
        return 0
    
    if (n == 0 or m == 0) and U == 0:
        return 1

    return num_seq(U - m, n - 1, m) + num_seq(U, n, m-1)


def dis_WMW_exact(U,n,m, lower_tail = True):
    '''
    Distribution function of the Wilcoxon rank sum statistic obtained from x, y with size n and m, respectively.

    Parameters
    ----------
    U : int
        Value of the Wilcoxon rank sum statistic for which the probability is calculated.
    n : int
        Sample size of x.
    m : int
        Sample size of y.

    Returns
    -------
    float
        Probability of Wilcoxon rank sum statistic smaller or equal than U

    '''
    # number of all possible combinations of x and y
    num_all = math.factorial(m + n) / math.factorial(m) / math.factorial(n)    

    # probability of Wilcoxon rank sum statistic smaller or equal than U
    prob = 0
    for k in np.arange(0,U+1): 
    
        num_res = num_seq(k, n, m)

        prob += num_res/num_all
        
    if lower_tail:
        
        return prob
    
    else:
        
        return 1 - prob
       



import numpy as np
from scipy.stats import rankdata


def check_ties(X, Y, ties):
    Z = np.concatenate((X, Y))

    # Observed samples in X and Y, respectively
    X_prime = X[~np.isnan(X)]
    Y_prime = Y[~np.isnan(Y)]

    # All observed samples
    Z_prime = np.concatenate((X_prime, Y_prime))
    r = rankdata(Z_prime)

    # Check ties
    if ties is None:
        # If ties is not specified, decide ties according to observed samples
        ties = len(r) != len(np.unique(r))
    else:
        # If ties is specified, consider false cases
        if not ties and len(r) != len(np.unique(r)):
            print("observed samples are tied, you may want to specify ties = True")

        if ties and len(r) == len(np.unique(r)) and len(Z) == len(Z_prime):
            print("all samples are observed distinct numbers, ties can only be False")
            ties = False

        if not ties and len(r) != len(np.unique(r)) and len(Z) == len(Z_prime):
            print("all samples are observed with tied observations, ties can only be True")
            ties = True

    return ties



import numpy as np
from scipy.stats import norm, rankdata
from .boundsWMWStatistic import bounds_wmw_statistic


def bounds_p_value_with_ties(X, Y, alternative, lower_boundary, upper_boundary, exact=None, correct=False):
    # Sample size
    n = len(X)
    m = len(Y)
    
    # Observed samples in X and Y, respectively
    X_prime = X[~np.isnan(X)]
    Y_prime = Y[~np.isnan(Y)]
    Z_prime = np.concatenate((X_prime, Y_prime))
    r = rankdata(Z_prime)
    
    # Observed sample size
    n_prime = len(X_prime)
    m_prime = len(Y_prime)
    
    # Compute bounds of WMWS statistic using boundsWMWStatistic function
    Bounds = bounds_wmw_statistic(X, Y, ties=True, lower_boundary=lower_boundary, upper_boundary=upper_boundary)
    lowerBoundWMWStatistic = Bounds[0]
    upperBoundWMWStatistic = Bounds[1]
    
    # Decide if exact
    if exact is None:
        exact = (n < 50) and (m < 50)
    
    if exact:
        print("Warning: cannot bound exact p-value with ties")
        exact = 0
    
    # Using normal approximation
    correct_lower = 0
    correct_upper = 0
    if correct:
        if alternative == 'two.sided':
            correct_lower = np.sign((lowerBoundWMWStatistic - n * m / 2)) * 0.5
            correct_upper = np.sign((upperBoundWMWStatistic - n * m / 2)) * 0.5
        elif alternative == 'greater':
            correct_lower = 0.5
            correct_upper = 0.5
        elif alternative == 'less':
            correct_lower = -0.5
            correct_upper = -0.5
    
    _, nties = np.unique(r, return_counts = True)
    
    mu = n * m / 2
    
    sigmaSquareMax = (n * m * (n + m + 1) / 12 -
                      n * m * np.sum(nties ** 3 - nties) / (12 * (n + m) * (n + m - 1)))
    
    nties[np.argmax(nties)] += n + m - n_prime - m_prime
    
    sigmaSquareMin = (n * m * (n + m + 1) / 12 -
                      n * m * np.sum(nties ** 3 - nties) / (12 * (n + m) * (n + m - 1)))
    
    if sigmaSquareMax == 0 or sigmaSquareMin == 0:
        raise ValueError("Cannot compute valid p-value")
    
    Z_1 = (lowerBoundWMWStatistic - correct_lower - mu) / np.sqrt(sigmaSquareMin)
    Z_2 = (upperBoundWMWStatistic - correct_upper - mu) / np.sqrt(sigmaSquareMax)
    Z_3 = (upperBoundWMWStatistic - correct_upper - mu) / np.sqrt(sigmaSquareMin)
    Z_4 = (lowerBoundWMWStatistic - correct_lower - mu) / np.sqrt(sigmaSquareMax)
    
    # Compute p-values
    if alternative == 'two.sided':
        p_1 = 2 * (1 - norm.cdf(np.abs(Z_1)))
        p_2 = 2 * (1 - norm.cdf(np.abs(Z_2)))
        p_3 = 2 * (1 - norm.cdf(np.abs(Z_3)))
        p_4 = 2 * (1 - norm.cdf(np.abs(Z_4)))
    elif alternative == 'greater':
        p_1 = 1 - norm.cdf(Z_1)
        p_2 = 1 - norm.cdf(Z_2)
        p_3 = 1 - norm.cdf(Z_3)
        p_4 = 1 - norm.cdf(Z_4)
    elif alternative == 'less':
        p_1 = norm.cdf(Z_1)
        p_2 = norm.cdf(Z_2)
        p_3 = norm.cdf(Z_3)
        p_4 = norm.cdf(Z_4)
    else:
        raise ValueError("Unknown alternative hypothesis")
    
    # Decide bounds of p-value
    if alternative == 'two.sided':
        if (lowerBoundWMWStatistic - n * m / 2 < 0) and (upperBoundWMWStatistic - n * m / 2 < 0):
            lowerBoundPValue = p_1
            upperBoundPValue = p_2
        elif (lowerBoundWMWStatistic - n * m / 2 < 0) and (upperBoundWMWStatistic - n * m / 2 >= 0):
            lowerBoundPValue = min(p_1, p_3)
            upperBoundPValue = 1
        else:
            lowerBoundPValue = p_3
            upperBoundPValue = p_4
    elif alternative == 'greater':
        if (lowerBoundWMWStatistic - n * m / 2 < 0) and (upperBoundWMWStatistic - n * m / 2 < 0):
            lowerBoundPValue = p_2
            upperBoundPValue = p_1
        elif (lowerBoundWMWStatistic - n * m / 2 < 0) and (upperBoundWMWStatistic - n * m / 2 >= 0):
            lowerBoundPValue = p_3
            upperBoundPValue = p_1
        else:
            lowerBoundPValue = p_3
            upperBoundPValue = p_4
    elif alternative == 'less':
        if (lowerBoundWMWStatistic - n * m / 2 < 0) and (upperBoundWMWStatistic - n * m / 2 < 0):
            lowerBoundPValue = p_1
            upperBoundPValue = p_2
        elif (lowerBoundWMWStatistic - n * m / 2 < 0) and (upperBoundWMWStatistic - n * m / 2 >= 0):
            lowerBoundPValue = p_1
            upperBoundPValue = p_3
        else:
            lowerBoundPValue = p_4
            upperBoundPValue = p_3
    else:
        raise ValueError("Unknown alternative hypothesis")
    
    return [lowerBoundWMWStatistic, upperBoundWMWStatistic, lowerBoundPValue, upperBoundPValue, exact]


# Example usage:
# X = np.array([1, 1, np.nan, 4, 5])
# Y = np.array([3, np.nan, 6, 7])
# alternative = 'two.sided'
# exact = None
# correct = True
# result = bounds_p_value_with_ties(X, Y, alternative, lower_boundary= 1, upper_boundary= np.inf, exact = exact, correct = correct)
# print("Lower bound of WMW Statistic:", result[0])
# print("Upper bound of WMW Statistic:", result[1])
# print("Lower bound of p-value:", result[2])
# print("Upper bound of p-value:", result[3])
# print("Exact:", result[4])



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="wmwm",
    version="0.0.1",
    author="Yijin Zeng",
    author_email="yijinzeng20@gmail.com",
    maintainer="Yijin Zeng",
    maintainer_email="yijinzeng20@gmail.com",
    description="A Python package performing Wilcoxon-Mann-Whitney test in the presence of missing data with controlled Type I error",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Yijin-Zeng/wmwm-python",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "numpy",
        "scipy"
    ],
    python_requires='>=3.6',
    license="MIT",
    keywords="Wilcoxon Mann-Whitney test, missing data, statistical analysis"
)



import numpy as np
from wmwm.checkTies import check_ties
from wmwm.boundsPValueWithTies import bounds_p_value_with_ties
from wmwm.boundsPValueNoTies import bounds_p_value_no_ties


def wmwm_test(X, Y, alternative='two.sided', ties=None,
              lower_boundary=-np.inf, upper_boundary=np.inf, exact=None, 
              correct=True):
    '''
    Performs the two-sample Wilcoxon-Mann-Whitney test in the presence of missing data, which controls the Type I error regardless of the values of missing data. For more details, see the reference (Zeng et al., 2024).
    
    Parameters
    ----------
    X,Y : numeric vectors of data values with potential missing data.
    alternative : a character string specifying the alternative hypothesis, must be one of "two.sided" (default), "greater" or "less".
    ties : a logical indicating whether samples could be tied. If observed samples contain tied samples, ties defaults to True. If observed samples do not contain tied samples, ties defaults to False.
    lower_boundary: (when ties is True) a number specifying the lower bound of the data set, must be smaller or equal than the minimum of all observed data.
    upper_boundary: (when ties is True) a number specifying the upper bound of the data set, must be larger or equal than the maximum of all observed data.
    exact: a logical indicating whether the bounds should be of an exact p-value.
    correct: a logical indicating whether the bounds should be of a p-value applying continuity correction in the normal approximation.

    Details
    -------
    wmwm_test performs the two-sample hypothesis test method proposed in (Zeng et al., 2024) for univariate data when not all data are observed. Bounds of the Wilcoxon-Mann-Whitney test statistic and its p-value will be computed in the presence of missing data. The p-value of the test method proposed in (Zeng et al., 2024) is then returned as the maximum possible p-value of the Wilcoxon-Mann-Whitney test.

    By default (if exact is not specified), this function returns bounds of an exact p-value if the X and Y both contain less than 50 samples and there are no ties. Otherwise, bounds of a p-value calculated using normal approximation with continuity correction will be returned.

        
    Returns
    -------
    p_value: the p-value for the test.
    bounds_statistic: bounds of the value of the Wilcoxon-Mann-Whitney test statistic.
    bounds_pvalue: bounds of the p-value of the Wilcoxon-Mann-Whitney test.
    alternative: a character string describing the alternative hypothesis.
    ties_method: a character string describing whether samples are considered tied.
    description_bounds: a character string describing the bounds of the p-value.

    
    Reference
    ----------
    [1] Mann HB, Whitney DR. On a test of whether one of two random variables is stochastically larger than the other. The annals of mathematical statistics. 1947 Mar 1:50-60.
    [2] Zeng Y, Adams NM, Bodenham DA. On two-sample testing for data with arbitrarily missing values. arXiv preprint arXiv:2403.15327. 2024 Mar 22.

    '''
    
    if alternative not in ('two.sided', 'less', 'greater'):
        
        raise ValueError("alternative must be one of 'two.sided', 'less', or 'greater'.")

    
    # Remove all infinite and NaN values
    X = X[np.isfinite(X) | np.isnan(X)]
    Y = Y[np.isfinite(Y) | np.isnan(Y)]

    # Check input
    if len(X) == 0 or len(Y) == 0:
        # Either X or Y does not contain any observed sample
        warning_msg = "either 'X' or 'Y' does not contain any observed sample"
        print(warning_msg)
        
        # Broadest bounds only
        BOUNDSWMW = np.array([0, len(X) * len(Y)])
        BOUNDSPVALUE = np.array([0, 1])
        DESCRIPTIONBOUNDS = "either 'X' or 'Y' does not contain any observed sample."
        
    else:
        # Both X and Y contain at least one observed sample
        ties = check_ties(X, Y, ties)
        
        # Compute bounds
        if ties:
            # Compute bounds of p-values with ties
            BOUNDS = bounds_p_value_with_ties(X, Y, alternative=alternative,
                                          lower_boundary=lower_boundary,
                                          upper_boundary=upper_boundary,
                                          exact=exact, correct=correct)
        else:
            # Compute bounds of p-values without ties
            BOUNDS = bounds_p_value_no_ties(X, Y, alternative=alternative,
                                        exact=exact, correct=correct)
        
        BOUNDSWMW = BOUNDS[:2]
        BOUNDSPVALUE = BOUNDS[2:4]
        exact = BOUNDS[4]
        
        # Description of bounds
        if exact:
            DESCRIPTIONBOUNDS = 'bounds_pvalue: bounds of the exact p-value'
        else:
            if correct:
                DESCRIPTIONBOUNDS = 'bounds_pvalue: bounds of the p-value obtained using normal approximation with continuity correction'
            else:
                DESCRIPTIONBOUNDS = 'bounds_pvalue: bounds of the p-value obtained using normal approximation without continuity correction'
    
    # Construct result dictionary
    RES = {
        'p_value': BOUNDSPVALUE[1],
        'bounds_statistic': BOUNDSWMW,
        'bounds_pvalue': BOUNDSPVALUE,
        'alternative': alternative,
        'ties_method': ties,
        'description_bounds': DESCRIPTIONBOUNDS
    }
    
    return RES
